package blusunrize.immersiveengineering.common.blocks.metal;

import blusunrize.immersiveengineering.api.ApiUtils;
import blusunrize.immersiveengineering.api.IEProperties;
import blusunrize.immersiveengineering.api.IEProperties.PropertyBoolInverted;
import blusunrize.immersiveengineering.api.TargetingInfo;
import blusunrize.immersiveengineering.api.energy.wires.IImmersiveConnectable;
import blusunrize.immersiveengineering.api.energy.wires.ImmersiveNetHandler.Connection;
import blusunrize.immersiveengineering.api.energy.wires.TileEntityImmersiveConnectable;
import blusunrize.immersiveengineering.api.energy.wires.WireType;
import blusunrize.immersiveengineering.common.blocks.IEBlockInterfaces.*;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumFacing.Axis;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.RayTraceResult;
import net.minecraft.util.math.Vec3d;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

public class TileEntityTransformer extends TileEntityImmersiveConnectable implements IDirectionalTile, IMirrorAble, IHasDummyBlocks, IAdvancedSelectionBounds, IDualState
{
	WireType secondCable;
	public EnumFacing facing=EnumFacing.NORTH;
	public int dummy=0;
	public boolean onPost=false;

	public static boolean _Immovable()
	{
		return true;
	}

	@Override
	protected boolean canTakeLV()
	{
		return true;
	}
	@Override
	protected boolean canTakeMV()
	{
		return true;
	}

	@Override
	public boolean canConnect()
	{
		return true;
	}
	//	@Override
	//	public boolean canUpdate()
	//	{
	//		return false;
	//	}
	@Override
	public boolean isEnergyOutput()
	{
		return false;
	}
	@Override
	public int outputEnergy(int amount, boolean simulate, int energyType)
	{
		return 0;
	}

	@Override
	public void writeCustomNBT(NBTTagCompound nbt, boolean descPacket)
	{
		super.writeCustomNBT(nbt, descPacket);
		nbt.func_74768_a("facing", facing.ordinal());
		if(secondCable!=null)
			nbt.func_74778_a("secondCable", secondCable.getUniqueName());
		nbt.func_74768_a("dummy", dummy);
		nbt.func_74757_a("postAttached", onPost);
	}
	@Override
	public void readCustomNBT(NBTTagCompound nbt, boolean descPacket)
	{
		super.readCustomNBT(nbt, descPacket);
		facing = EnumFacing.func_82600_a(nbt.func_74762_e("facing"));
		if(nbt.func_74764_b("secondCable"))
			secondCable = ApiUtils.getWireTypeFromNBT(nbt, "secondCable");
		else
			secondCable = null;
		dummy = nbt.func_74762_e("dummy");
		onPost = nbt.func_74767_n("postAttached");
	}

	@Override
	public BlockPos getConnectionMaster(WireType cableType, TargetingInfo target)
	{
		return func_174877_v().func_177982_a(0,-dummy,0);
	}	
	@Override
	public boolean canConnectCable(WireType cableType, TargetingInfo target)
	{
		if(cableType==WireType.STEEL&&!canTakeHV())
			return false;
		if(dummy!=0)
		{
			TileEntity master = field_145850_b.func_175625_s(func_174877_v().func_177982_a(0,-dummy,0));
			if(master instanceof TileEntityTransformer)
				return ((TileEntityTransformer)master).canConnectCable(cableType, target);
			return false;
		}
		int tc = getTargetedConnector(target);
		switch(tc)
		{
		case 0:
			return limitType==null && secondCable!=cableType;
		case 1:
			return secondCable==null && limitType!=cableType;
		}
		return false;
	}
	@Override
	public void connectCable(WireType cableType, TargetingInfo target, IImmersiveConnectable other)
	{
		if(dummy!=0)
		{
			TileEntity master = field_145850_b.func_175625_s(func_174877_v().func_177982_a(0,-dummy,0));
			if(master instanceof TileEntityTransformer)
				((TileEntityTransformer) master).connectCable(cableType, target, other);
			return;
		}
		switch(getTargetedConnector(target))
		{
		case 0:
			if(this.limitType==null)
				this.limitType = cableType;
			break;
		case 1:
			if(secondCable==null)
				this.secondCable = cableType;
			break;
		}
	}
	@Override
	public WireType getCableLimiter(TargetingInfo target)
	{
		switch(getTargetedConnector(target))
		{
		case 0:
			return limitType;
		case 1:
			return secondCable;
		}
		return null;
	}
	@Override
	public void removeCable(Connection connection)
	{
		WireType type = connection!=null?connection.cableType:null;
		if(type==null)
		{
			limitType=null;
			secondCable=null;
		}
		if(type==limitType)
			this.limitType = null;
		if(type==secondCable)
			this.secondCable = null;
		this.markContainingBlockForUpdate(null);
	}

	@Override
	public Vec3d getRaytraceOffset(IImmersiveConnectable link)
	{
		if(onPost)
			return new Vec3d(.5, 1.5, .5);
		return new Vec3d(.5, 2.75, .5);
	}
	@Override
	public Vec3d getConnectionOffset(Connection con)
	{
		boolean b = con.cableType==limitType;
		if(onPost)
		{
			if(b)
				return new Vec3d(.5+(facing==EnumFacing.EAST?.4375: facing==EnumFacing.WEST?-.4375: 0),1.4375,.5+(facing==EnumFacing.SOUTH?.4375: facing==EnumFacing.NORTH?-.4375: 0));
			else
				return new Vec3d(.5+(facing==EnumFacing.EAST?-.0625: facing==EnumFacing.WEST?.0625: 0),.25,.5+(facing==EnumFacing.SOUTH?-.0625: facing==EnumFacing.NORTH?.0625: 0));
		}
		else
		{
			double conRadius = con.cableType.getRenderDiameter()/2;
			double offset = con.cableType==WireType.COPPER?-.0625: con.cableType==WireType.ELECTRUM?.0625: .25; 
			if(facing==EnumFacing.NORTH)
				return new Vec3d(b?.8125:.1875, 2.5+offset-conRadius, .5);
			if(facing==EnumFacing.SOUTH)
				return new Vec3d(b?.1875:.8125, 2.5+offset-conRadius, .5);
			if(facing==EnumFacing.WEST)
				return new Vec3d(.5, 2.5+offset-conRadius, b?.1875:.8125);
			if(facing==EnumFacing.EAST)
				return new Vec3d(.5, 2.5+offset-conRadius, b?.8125:.1875);
		}
		return new Vec3d(.5,.5,.5);
	}

	public int getTargetedConnector(TargetingInfo target)
	{
		if(onPost)
		{
			if(target.hitY>=.5)
				return 0;
			else
				return 1;
		}
		else
		{
			if(facing==EnumFacing.NORTH)
				if(target.hitX<.5)
					return 1;
				else
					return 0;
			else if(facing==EnumFacing.SOUTH)
				if(target.hitX<.5)
					return 0;
				else
					return 1;
			else if(facing==EnumFacing.WEST)
				if(target.hitZ<.5)
					return 0;
				else
					return 1;
			else if(facing==EnumFacing.EAST)
				if(target.hitZ<.5)
					return 1;
				else
					return 0;
		}
		return -1;
	}

	public WireType getLimiter(int side)
	{
		if(side==0)
			return limitType;
		return secondCable;
	}

	@SideOnly(Side.CLIENT)
	private AxisAlignedBB renderAABB;
	@SideOnly(Side.CLIENT)
	@Override
	public AxisAlignedBB getRenderBoundingBox()
	{
		//		if(renderAABB==null)
		//		{
		//			if(Config.getBoolean("increasedRenderboxes"))
		//				renderAABB = AxisAlignedBB.getBoundingBox(xCoord-16, yCoord-16, zCoord-16, xCoord+17, yCoord+17, zCoord+17);
		//			else if(!dummy)
		//				renderAABB = AxisAlignedBB.getBoundingBox(xCoord, yCoord-1, zCoord, xCoord+1, yCoord+1.5, zCoord+1);
		//			else
		//				renderAABB = AxisAlignedBB.getBoundingBox(xCoord, yCoord, zCoord, xCoord, yCoord, zCoord);
		//		}
		return new AxisAlignedBB(func_174877_v().func_177958_n()-16,func_174877_v().func_177956_o()-16,func_174877_v().func_177952_p()-16, func_174877_v().func_177958_n()+16,func_174877_v().func_177956_o()+16,func_174877_v().func_177952_p()+16);
	}

	@Override
	public PropertyBoolInverted getBoolProperty(Class<? extends IUsesBooleanProperty> inf)
	{
		return inf==IMirrorAble.class?IEProperties.BOOLEANS[0]:inf==IDualState.class?IEProperties.BOOLEANS[1]:null;
	}
	@Override
	public boolean getIsMirrored()
	{
		if (onPost)
			return false;
		WireType lower = this instanceof TileEntityTransformerHV?WireType.ELECTRUM:WireType.COPPER;
		boolean b = (limitType!=null&&lower.equals(limitType))||(secondCable!=null&&!lower.equals(secondCable));
		return !b;
	}

	@Override
	public EnumFacing getFacing()
	{
		return facing;
	}
	@Override
	public void setFacing(EnumFacing facing)
	{
		this.facing = facing;
	}
	@Override
	public int getFacingLimitation()
	{
		return 2;
	}
	@Override
	public boolean mirrorFacingOnPlacement(EntityLivingBase placer)
	{
		return false;
	}
	@Override
	public boolean canHammerRotate(EnumFacing side, float hitX, float hitY, float hitZ, EntityLivingBase entity)
	{
		return false;
	}
	@Override
	public boolean canRotate(EnumFacing axis)
	{
		return false;
	}

	@Override
	public boolean isDummy()
	{
		return dummy!=0;
	}
	public boolean isOnPost()
	{
		return onPost;
	}
	@Override
	public void placeDummies(BlockPos pos, IBlockState state, EnumFacing side, float hitX, float hitY, float hitZ)
	{
		if (state.func_177229_b(IEProperties.BOOLEANS[1]))
		{
			onPost = true;
			func_70296_d();
			this.markContainingBlockForUpdate(null);
		}
		else
			for(int i=1; i<=2; i++)
			{
				field_145850_b.func_175656_a(pos.func_177982_a(0,i,0), state);
				((TileEntityTransformer)field_145850_b.func_175625_s(pos.func_177982_a(0,i,0))).dummy = i;
				((TileEntityTransformer)field_145850_b.func_175625_s(pos.func_177982_a(0,i,0))).facing = this.facing;
			}
	}
	@Override
	public void breakDummies(BlockPos pos, IBlockState state)
	{
		if (onPost)
			return;
		for(int i=0; i<=2; i++)
			field_145850_b.func_175698_g(func_174877_v().func_177982_a(0,-dummy,0).func_177982_a(0,i,0));
	}

	@Override
	public float[] getBlockBounds()
	{
		if(dummy==2)
			return new float[]{facing.func_176740_k()==Axis.Z?0:.3125f, 0, facing.func_176740_k()==Axis.X?0:.3125f, facing.func_176740_k()==Axis.Z?1:.6875f,this instanceof TileEntityTransformerHV?.75f:.5625f,facing.func_176740_k()==Axis.X?1:.6875f};
		if (onPost)
			return new float[]{facing.func_176740_k()==Axis.Z?.25F:facing==EnumFacing.WEST?-.375F:.6875F,
					0,
					facing.func_176740_k()==Axis.X?.25F:facing==EnumFacing.NORTH?-.375F:.6875F,
					facing.func_176740_k()==Axis.Z?.75F:facing==EnumFacing.EAST?1.375F:.3125F,
					1,
					facing.func_176740_k()==Axis.X?.75F:facing==EnumFacing.SOUTH?1.375F:.3125F};
		return null;
	}

	@Override
	public List<AxisAlignedBB> getAdvancedSelectionBounds()
	{
		if(dummy==2)
		{
			WireType lower = this instanceof TileEntityTransformerHV?WireType.ELECTRUM:WireType.COPPER;
			WireType higher = this instanceof TileEntityTransformerHV?WireType.STEEL:WireType.ELECTRUM;
			boolean b = (limitType!=null&&lower.equals(limitType))||(secondCable!=null&&!lower.equals(secondCable));
			double offsetL = lower==WireType.COPPER?0: lower==WireType.ELECTRUM?.0625: .25; 
			double offsetH = higher==WireType.COPPER?0: higher==WireType.ELECTRUM?.0625: .25; 
			if(facing==EnumFacing.NORTH)
				return Lists.newArrayList(new AxisAlignedBB(0,0,.3125, .375,.5+(b?offsetH:offsetL),.6875).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()), new AxisAlignedBB(.625,0,.3125, 1,.5+(b?offsetL:offsetH),.6875).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()));
			if(facing==EnumFacing.SOUTH)
				return Lists.newArrayList(new AxisAlignedBB(0,0,.3125, .375,.5+(b?offsetL:offsetH),.6875).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()), new AxisAlignedBB(.625,0,.3125, 1,.5+(b?offsetH:offsetL),.6875).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()));
			if(facing==EnumFacing.WEST)
				return Lists.newArrayList(new AxisAlignedBB(.3125,0,0, .6875,.5+(b?offsetL:offsetH),.375).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()), new AxisAlignedBB(.3125,0,.625, .6875,.5+(b?offsetH:offsetL),1).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()));
			if(facing==EnumFacing.EAST)
				return Lists.newArrayList(new AxisAlignedBB(.3125,0,0, .6875,.5+(b?offsetH:offsetL),.375).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()), new AxisAlignedBB(.3125,0,.625, .6875,.5+(b?offsetL:offsetH),1).func_72317_d(func_174877_v().func_177958_n(),func_174877_v().func_177956_o(),func_174877_v().func_177952_p()));
		}
		return null;
	}

	@Override
	public boolean isOverrideBox(AxisAlignedBB box, EntityPlayer player, RayTraceResult mop, ArrayList<AxisAlignedBB> list)
	{
		return box.func_72314_b(.002, .002, .002).func_72318_a(mop.field_72307_f);
	}
	@Override
	public Set<BlockPos> getIgnored(IImmersiveConnectable other)
	{
		if (onPost)
			return super.getIgnored(other);
		return ImmutableSet.of(field_174879_c.func_177984_a(), field_174879_c.func_177981_b(2));
	}

	@Override
	public boolean getIsSecondState()
	{
		return onPost;
	}
}
